import {PersonExtended} from "@admin/models/person-extended.model";
import {PersonRole} from "@admin/models/person-role.model";
import {Relation3TiersPersonOrgunitRole} from "@admin/models/relation-3-tiers-person-orgunit-role.model";
import {
  ChangeDetectionStrategy,
  ChangeDetectorRef,
  Component,
  HostBinding,
  Input,
  OnInit,
  Output,
} from "@angular/core";
import {
  AbstractControl,
  ControlValueAccessor,
  FormArray,
  FormBuilder,
  FormControl,
  FormGroup,
  NG_VALUE_ACCESSOR,
  Validators,
} from "@angular/forms";
import {
  OrganizationalUnit,
  Role,
} from "@app/generated-api";
import {SharedAbstractPresentational} from "@app/shared/components/presentationals/shared-abstract/shared-abstract.presentational";
import {TranslateService} from "@ngx-translate/core";
import {Store} from "@ngxs/store";
import {RoleEnum} from "@shared/enums/role.enum";
import {RoutesEnum} from "@shared/enums/routes.enum";
import {BaseFormDefinition} from "@shared/models/base-form-definition.model";
import {
  BehaviorSubject,
  Observable,
} from "rxjs";
import {tap} from "rxjs/operators";
import {
  FormValidationHelper,
  isEmptyArray,
  isNullOrUndefined,
  isUndefined,
  ObservableUtil,
  PropertyName,
  StringUtil,
  TRANSLATE,
} from "solidify-frontend";

@Component({
  selector: "dlcm-shared-person-orgunit-role",
  templateUrl: "./shared-person-orgunit-role.presentational.html",
  styleUrls: ["./shared-person-orgunit-role.presentational.scss"],
  changeDetection: ChangeDetectionStrategy.OnPush,
  providers: [
    {
      provide: NG_VALUE_ACCESSOR,
      multi: true,
      useExisting: SharedPersonOrgunitRolePresentational,
    },
  ],
})
export class SharedPersonOrgunitRolePresentational extends SharedAbstractPresentational implements ControlValueAccessor, OnInit {
  formDefinition: FormComponentFormDefinition = new FormComponentFormDefinition();
  formArray: FormArray;

  @Input()
  mode: PersonOrgUnitRoleMode = PersonOrgUnitRoleMode.undefined;

  @HostBinding("class.disabled") disabled: boolean = false;

  @Input()
  selectedPersonOrOrgUnitRole: Relation3TiersPersonOrgunitRole[];

  @Input()
  listPerson: PersonExtended[];

  @Input()
  listOrgUnit: OrganizationalUnit[];

  @Input()
  listRole: Role[];

  @Input()
  formControl: FormControl;

  _readonly: boolean;

  @Input()
  set readonly(value: boolean) {
    this._readonly = value;
    this.updateFormReadonlyState();
  }

  get readonly(): boolean {
    return this._readonly;
  }

  private readonly _valueBS: BehaviorSubject<any[] | undefined> = new BehaviorSubject<any[] | undefined>(undefined);

  @Output("valueChange")
  readonly valueObs: Observable<any[] | undefined> = ObservableUtil.asObservable(this._valueBS);

  protected readonly _navigateBS: BehaviorSubject<string[]> = new BehaviorSubject<string[]>(undefined);
  @Output("navigate")
  readonly navigateObs: Observable<string[]> = ObservableUtil.asObservable(this._navigateBS);

  get modeEnum(): typeof PersonOrgUnitRoleMode {
    return PersonOrgUnitRoleMode;
  }

  get formValidationHelper(): typeof FormValidationHelper {
    return FormValidationHelper;
  }

  getFormControl(formControl: AbstractControl, key: string): AbstractControl {
    return formControl.get(key);
  }

  isRequired(formControl: AbstractControl, key: string): boolean {
    const errors = formControl.get(key).errors;
    return isNullOrUndefined(errors) ? false : errors.required;
  }

  constructor(private translate: TranslateService,
              private store: Store,
              private readonly _fb: FormBuilder,
              private readonly _changeDetector: ChangeDetectorRef) {
    super();
  }

  ngOnInit(): void {
    super.ngOnInit();

    // TODO Manage better form to use directly formControl (allow to share disable and valid state)
    this.formArray = this._fb.array([]);

    if (!isNullOrUndefined(this.selectedPersonOrOrgUnitRole)) {
      this.selectedPersonOrOrgUnitRole.forEach(p => {
        this.formArray.push(this.createPersonOrOrgunitRole(p));
      });
    }

    this.updateFormReadonlyState();
    this.formControl.setValue(this.formArray.value);

    this.subscribe(this.formArray.valueChanges.pipe(
      tap(value => {
        this.propagateChange(value);
        this._valueBS.next(value);
      }),
    ));
  }

  private updateFormReadonlyState(): void {
    if (isNullOrUndefined(this.formArray) || isNullOrUndefined(this.formControl)) {
      return;
    }
    if (this._readonly) {
      this.disabled = true;
      this.formArray.disable();
      this.formControl.disable();
    } else {
      this.disabled = false;
      this.formArray.enable();
      this.formControl.enable();
    }
  }

  createPersonOrOrgunitRole(person: PersonRole | undefined = undefined): FormGroup {

    if (isUndefined(person)) {
      const form = this._fb.group({
        [this.formDefinition.id]: ["", [Validators.required]],
        [this.formDefinition.listId]: ["", [Validators.required]],
      });
      return form;
    } else {
      // We transform here list of roles into unique role because we hide the multiples roles to the final user
      const roleId = person.roles.length > 0 ? this.getMostImportantRole(person.roles.map(r => r.resId) as RoleEnum[]) : "";
      const form = this._fb.group({
        [this.formDefinition.id]: [person.resId, [Validators.required]],
        [this.formDefinition.listId]: [roleId, [Validators.required]],
      });
      if (person.isReadonly) {
        setTimeout(() => {
          form.disable();
        }, 0);
      }
      return form;
    }
  }

  addPersonOrOrgUnit(): void {
    this.formArray.push(this.createPersonOrOrgunitRole());
  }

  delete(index: number): void {
    this.formArray.removeAt(index);
  }

  filterPerson(formControl: AbstractControl, listPerson: PersonExtended[]): PersonExtended[] {
    return this.filter(listPerson, formControl);
  }

  filterOrgUnit(formControl: AbstractControl, listOrgUnit: OrganizationalUnit[]): OrganizationalUnit[] {
    return this.filter(listOrgUnit, formControl);
  }

  private filter(listPersonOrOrgUnit: OrganizationalUnit[], formControl: AbstractControl): any {
    if (isNullOrUndefined(listPersonOrOrgUnit) || isEmptyArray(listPersonOrOrgUnit)) {
      return listPersonOrOrgUnit;
    }
    const personOrOrgUnitAlreadySelected = this.formArray.controls.map(c => c.get(this.formDefinition.id).value);
    return listPersonOrOrgUnit.filter(p => p.resId === formControl.value || !personOrOrgUnitAlreadySelected.includes(p.resId));
  }

  propagateChange = (__: any) => {};

  registerOnChange(fn: any): void {
    this.propagateChange = fn;
  }

  registerOnTouched(fn: any): void {
  }

  setDisabledState(isDisabled: boolean): void {
  }

  writeValue(value: string[]): void {
    // if (value) {
    // this.formControl.setValue(value);
    // }
  }

  getTitle(): string {
    if (this.mode === PersonOrgUnitRoleMode.person) {
      return TRANSLATE("shared.personOrOrgUnitRole.title.person");
    }
    if (this.mode === PersonOrgUnitRoleMode.orgUnit) {
      return TRANSLATE("shared.personOrOrgUnitRole.title.orgUnit");
    }
  }

  getMessageNoItem(): string {
    if (this.mode === PersonOrgUnitRoleMode.person) {
      return TRANSLATE("shared.personOrOrgUnitRole.noItem.person");
    }
    if (this.mode === PersonOrgUnitRoleMode.orgUnit) {
      return TRANSLATE("shared.personOrOrgUnitRole.noItem.orgUnit");
    }
  }

  private getMostImportantRole(listRoles: RoleEnum[]): RoleEnum | string {
    if (listRoles.length === 0) {
      return "";
    }
    if (listRoles.length === 1) {
      return listRoles[0];
    }
    let role = listRoles.find(r => r === RoleEnum.MANAGER);
    if (!isNullOrUndefined(role)) {
      return RoleEnum.MANAGER;
    }
    role = listRoles.find(r => r === RoleEnum.STEWARD);
    if (!isNullOrUndefined(role)) {
      return RoleEnum.STEWARD;
    }
    role = listRoles.find(r => r === RoleEnum.APPROVER);
    if (!isNullOrUndefined(role)) {
      return RoleEnum.APPROVER;
    }
    role = listRoles.find(r => r === RoleEnum.CREATOR);
    if (!isNullOrUndefined(role)) {
      return RoleEnum.CREATOR;
    }
    role = listRoles.find(r => r === RoleEnum.VISITOR);
    if (!isNullOrUndefined(role)) {
      return RoleEnum.VISITOR;
    }
    return StringUtil.stringEmpty;
  }

  goToPerson(personId: string): void {
    this._navigateBS.next([RoutesEnum.adminPersonDetail, personId]);
  }

  goToOrgUnit(orgUnitId: string): void {
    this._navigateBS.next([RoutesEnum.adminOrganizationalUnitDetail, orgUnitId]);
  }
}

class FormComponentFormDefinition extends BaseFormDefinition {
  @PropertyName() id: string;
  @PropertyName() listId: string;
}

export enum PersonOrgUnitRoleMode {
  undefined,
  person,
  orgUnit,
}
